/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Netscape Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s): Radha Kulkarni(radha@netscape.com)
 */

#include "nsWyciwygChannel.h"
#include "nsIServiceManager.h"
#include "nsILoadGroup.h"
#include "nsIScriptSecurityManager.h"
#include "nsNetUtil.h"
#include "nsICacheService.h"
#include "nsICacheSession.h"


PRLogModuleInfo * gWyciwygLog = nsnull;

#define wyciwyg_TYPE "text/html"

// nsWyciwygChannel methods 
nsWyciwygChannel::nsWyciwygChannel()
    : mLoadAttributes(nsIChannel::LOAD_NORMAL),
      mStatus(NS_OK),
      mIsPending(PR_FALSE),
      mLoadFlags(LOAD_NORMAL)
{
    NS_INIT_REFCNT();
}

nsWyciwygChannel::~nsWyciwygChannel() {
}

NS_IMPL_THREADSAFE_ISUPPORTS7(nsWyciwygChannel, nsIChannel, nsIRequest,
                              nsIStreamListener, nsICacheListener, 
                              nsIInterfaceRequestor, nsIWyciwygChannel,
                              nsIRequestObserver)

nsresult
nsWyciwygChannel::Init(nsIURI* uri)
{
    if (!uri)
        return NS_ERROR_NULL_POINTER;
    mURI = uri;
    return NS_OK;
}

NS_METHOD
nsWyciwygChannel::Create(nsISupports* aOuter, const nsIID& aIID, void* *aResult)
{
    nsWyciwygChannel* wc = new nsWyciwygChannel();
    if (wc == nsnull)
        return NS_ERROR_OUT_OF_MEMORY;
    NS_ADDREF(wc);
    nsresult rv = wc->QueryInterface(aIID, aResult);
    NS_RELEASE(wc);
    return rv;
}

//-----------------------------------------------------------------------------
// nsHttpChannel::nsIInterfaceRequestor
//-----------------------------------------------------------------------------

NS_IMETHODIMP
nsWyciwygChannel::GetInterface(const nsIID &iid, void **result)
{
  return QueryInterface(iid, result);

  // XXX Should I do this?
#if 0
  if (mCallbacks)
        return mCallbacks->GetInterface(iid, result);

    return NS_ERROR_NO_INTERFACE;
#endif
}

///////////////////////////////////////////////////////////////////////////////
// nsIRequest methods:
///////////////////////////////////////////////////////////////////////////////

NS_IMETHODIMP
nsWyciwygChannel::GetName(PRUnichar**aName)
{
  NS_ENSURE_ARG_POINTER(aName);
  nsXPIDLCString spec;
  mURI->GetSpec(getter_Copies(spec));
  *aName = ToNewUnicode(spec);
  return NS_OK;
}
 
NS_IMETHODIMP
nsWyciwygChannel::IsPending(PRBool *aIsPending)
{
  *aIsPending = mIsPending;
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::GetStatus(nsresult *aStatus)
{
    *aStatus = mStatus;
    return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::Cancel(nsresult aStatus)
{
  LOG(("nsWyciwygChannel::Cancel [this=%x status=%x]\n", this, aStatus));
  NS_ASSERTION(NS_FAILED(aStatus), "shouldn't cancel with a success code");
  
  mStatus = aStatus;
  if (mCacheReadRequest)
    mCacheReadRequest->Cancel(aStatus);
  // Clear out all cache handles.
  CloseCacheEntry();
  return NS_OK;
}
 
NS_IMETHODIMP
nsWyciwygChannel::Suspend(void)
{
  LOG(("nsWyciwygChannel::Suspend [this=%x]\n", this));
  if (mCacheReadRequest)
    return mCacheReadRequest->Suspend();
  return NS_OK;
}
 
NS_IMETHODIMP
nsWyciwygChannel::Resume(void)
{
  LOG(("nsWyciwygChannel::Resume [this=%x]\n", this));
  if (mCacheReadRequest)
    return mCacheReadRequest->Resume();
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::GetLoadGroup(nsILoadGroup* *aLoadGroup)
{
  NS_ENSURE_ARG_POINTER(aLoadGroup);
  *aLoadGroup = mLoadGroup;
  NS_IF_ADDREF(*aLoadGroup);
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::SetLoadGroup(nsILoadGroup* aLoadGroup)
{
  mLoadGroup = aLoadGroup;
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::SetLoadFlags(PRUint32 aLoadFlags)
{
  mLoadFlags = aLoadFlags;
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::GetLoadFlags(PRUint32 * aLoadFlags)
{
  NS_ENSURE_ARG_POINTER(aLoadFlags);
  *aLoadFlags = mLoadFlags;
  return NS_OK;
}

////////////////////////////////////////////////////////////////////////////////
// nsIChannel methods:
///////////////////////////////////////////////////////////////////////////////
NS_IMETHODIMP
nsWyciwygChannel::GetOriginalURI(nsIURI* *aURI)
{
  // Let's hope this isn't called before mOriginalURI is set or we will
  // return the full wyciwyg URI for our originalURI  :S
  NS_ASSERTION(mOriginalURI, "nsWyciwygChannel::GetOriginalURI - mOriginalURI not set!\n");

  *aURI = mOriginalURI ? mOriginalURI : mURI;
  NS_IF_ADDREF(*aURI);
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::SetOriginalURI(nsIURI* aURI)
{
  mOriginalURI = aURI;
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::GetURI(nsIURI* *aURI)
{
  NS_ENSURE_ARG_POINTER(aURI);
  *aURI = mURI;
  NS_IF_ADDREF(*aURI);
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::GetOwner(nsISupports* *aOwner)
{
  nsresult rv = NS_OK;
  if (!mOwner) {
    // Create codebase principal with URI of original document, not our URI
    NS_ASSERTION(mOriginalURI,
        "nsWyciwygChannel::GetOwner without an owner or an original URI!");
    if (mOriginalURI) {
      nsIPrincipal* pIPrincipal = nsnull;
      nsCOMPtr<nsIScriptSecurityManager> secMan(do_GetService(NS_SCRIPTSECURITYMANAGER_CONTRACTID, &rv));
      if (secMan) {
        rv = secMan->GetCodebasePrincipal(mOriginalURI, &pIPrincipal);
        if (NS_SUCCEEDED(rv)) {
          mOwner = pIPrincipal;
          NS_RELEASE(pIPrincipal);
        }
      }
    } else {
      // Uh oh, must set originalURI before we can return an owner!
      return NS_ERROR_FAILURE;
    }
  }
  NS_ASSERTION(mOriginalURI,
      "nsWyciwygChannel::GetOwner unable to get owner!");
  if (mOwner) {
    *aOwner = mOwner.get();
    NS_IF_ADDREF(*aOwner);
  } else {
    *aOwner = nsnull;
  }
  return rv;
}

NS_IMETHODIMP
nsWyciwygChannel::SetOwner(nsISupports* aOwner)
{
  mOwner = aOwner;
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::GetNotificationCallbacks(nsIInterfaceRequestor* *aCallbacks)
{
  *aCallbacks = mCallbacks.get();
  NS_IF_ADDREF(*aCallbacks);
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::SetNotificationCallbacks(nsIInterfaceRequestor* aNotificationCallbacks)
{
  mCallbacks = aNotificationCallbacks;
  return NS_OK;
}

NS_IMETHODIMP 
nsWyciwygChannel::GetSecurityInfo(nsISupports * *aSecurityInfo)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsWyciwygChannel::GetContentType(char* *aContentType)
{
  NS_ENSURE_ARG_POINTER(aContentType);
  *aContentType = nsCRT::strdup(wyciwyg_TYPE);
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::SetContentType(const char *aContentType)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsWyciwygChannel::GetContentLength(PRInt32 *aContentLength)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsWyciwygChannel::SetContentLength(PRInt32 aContentLength)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsWyciwygChannel::Open(nsIInputStream ** aReturn)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsWyciwygChannel::AsyncOpen(nsIStreamListener * aListener, nsISupports  * aContext)
{
    LOG(("nsWyciwygChannel::AsyncOpen [this=%x]\n", this));

    NS_ENSURE_ARG_POINTER(aListener);
    NS_ENSURE_TRUE(!mIsPending, NS_ERROR_IN_PROGRESS);

    //XXX Should I worry about Port safety? 

    mIsPending = PR_TRUE;
    mListener = aListener;
    mListenerContext = aContext;

    // add ourselves to the load group. From this point forward, we'll report
    // all failures asynchronously.
    if (mLoadGroup)
        mLoadGroup->AddRequest(this, nsnull);

    nsresult rv = Connect(PR_TRUE);
    if (NS_FAILED(rv)) {
        LOG(("nsWyciwygChannel::AsyncOpen Connect failed [rv=%x]\n", rv));
        CloseCacheEntry();
        AsyncAbort(rv);
    }
    return NS_OK;
}

//////////////////////////////////////////////////////////////////////////////
// nsIWyciwygChannel
//////////////////////////////////////////////////////////////////////////////

NS_IMETHODIMP
nsWyciwygChannel::CreateCacheEntry(const char * aCacheKey)
{
  return OpenCacheEntry(aCacheKey, PR_TRUE);
}

NS_IMETHODIMP
nsWyciwygChannel::WriteToCache(const char * aScript)
{
  if (!mCacheEntry)
    return NS_ERROR_FAILURE;
  PRUint32 len = nsCRT::strlen(aScript);
  nsresult rv;

  if (!mCacheTransport && !mCacheOutputStream) {
    //Get the transport from cache
    rv = mCacheEntry->GetTransport(getter_AddRefs(mCacheTransport)); 
  
    // Get the outputstream from the transport.
    if (mCacheTransport)
      rv = mCacheTransport->OpenOutputStream(0, -1, 0, getter_AddRefs(mCacheOutputStream));    
  }

  if (mCacheOutputStream)
    rv = mCacheOutputStream->Write(aScript, len, &out);
  return rv;
}


NS_IMETHODIMP
nsWyciwygChannel::CloseCacheEntry()
{
  nsresult rv = NS_OK;
  if (mCacheEntry) {
    LOG(("nsWyciwygChannel::CloseCacheEntry [this=%x ]", this));
    // make sure the cache transport isn't holding a reference back to us
    if (mCacheTransport)
      mCacheTransport->SetNotificationCallbacks(nsnull, 0);
    mCacheReadRequest = 0;
    mCacheTransport = 0;
    mCacheOutputStream = 0;
    mCacheEntry = 0;
  }
  return rv;
}

//////////////////////////////////////////////////////////////////////////////
// nsICachelistener
//////////////////////////////////////////////////////////////////////////////
NS_IMETHODIMP
nsWyciwygChannel::OnCacheEntryAvailable(nsICacheEntryDescriptor * aCacheEntry, nsCacheAccessMode aMode, nsresult aStatus)
{
      LOG(("nsWyciwygChannel::OnCacheEntryAvailable [this=%x entry=%x "
         "access=%x status=%x]\n", this, aCacheEntry, aMode, aStatus));

    // if the channel's already fired onStopRequest, 
    // then we should ignore this event.
    if (!mIsPending)
        return NS_OK;

    // otherwise, we have to handle this event.
    if (NS_SUCCEEDED(aStatus)) {
        mCacheEntry = aCacheEntry;        
    }

    nsresult rv;

    if (NS_FAILED(mStatus)) {
        LOG(("channel was canceled [this=%x status=%x]\n", this, mStatus));
        rv = mStatus;
    }
    else // advance to the next state...
        rv = Connect(PR_FALSE);

    // a failure from Connect means that we have to abort the channel.
    if (NS_FAILED(rv)) {
        CloseCacheEntry();
        AsyncAbort(rv);
    }

    return rv;
}

//-----------------------------------------------------------------------------
// nsWyciwygChannel::nsIStreamListener
//-----------------------------------------------------------------------------

NS_IMETHODIMP
nsWyciwygChannel::OnDataAvailable(nsIRequest *request, nsISupports *ctxt,
                               nsIInputStream *input,
                               PRUint32 offset, PRUint32 count)
{
    LOG(("nsWyciwygChannel::OnDataAvailable [this=%x request=%x offset=%u count=%u]\n",
        this, request, offset, count));

    // if the request is for something we no longer reference, then simply 
    // drop this event.
    if (request != mCacheReadRequest) {
        NS_WARNING("nsWyciwygChannel::OnDataAvailable got stale request... why wasn't it cancelled?");
        return NS_BASE_STREAM_CLOSED;
    }

    if (mListener)
        return mListener->OnDataAvailable((nsIRequest *)this, mListenerContext, input, offset, count);

    return NS_BASE_STREAM_CLOSED;
}

//////////////////////////////////////////////////////////////////////////////
// nsIRequestObserver
//////////////////////////////////////////////////////////////////////////////

NS_IMETHODIMP
nsWyciwygChannel::OnStartRequest(nsIRequest *request, nsISupports *ctxt)
{
  nsresult rv = NS_ERROR_FAILURE;
  LOG(("nsWyciwygChannel::OnStartRequest [this=%x request=%x\n",
        this, request));

  // capture the request's status, so our consumers will know ASAP of any
  // connection failures, etc.
  request->GetStatus(&mStatus);
  if (mListener)
     rv = mListener->OnStartRequest(this, mListenerContext);
  return rv;
}


NS_IMETHODIMP
nsWyciwygChannel::OnStopRequest(nsIRequest *request, nsISupports *ctxt, nsresult status)
{
  nsresult rv = NS_ERROR_FAILURE;
  LOG(("nsWyciwygChannel::OnStopRequest [this=%x request=%x status=%d\n",
        this, request, (PRUint32)status));
  
  mIsPending = PR_FALSE;
  mStatus = status;
  CloseCacheEntry();
  if (mListener) {
    rv = mListener->OnStopRequest(this, mListenerContext, status);
     mListener = 0;
     mListenerContext = 0;
  }
  
  if (mLoadGroup)
    rv = mLoadGroup->RemoveRequest(this, nsnull, status);

  return NS_OK;

}

//////////////////////////////////////////////////////////////////////////////
// Helper functions
//////////////////////////////////////////////////////////////////////////////
nsresult
nsWyciwygChannel::OpenCacheEntry(const char * aCacheKey, PRBool aWriteAccess, PRBool * aDelayFlag )
{
  nsresult rv;
  // Get cache service
  nsCOMPtr<nsICacheService>  cacheService(do_GetService(NS_CACHESERVICE_CONTRACTID, &rv));

  if (NS_SUCCEEDED(rv) && cacheService) {
    nsXPIDLCString spec;    
    nsAutoString newURIString;    
    nsCOMPtr<nsICacheSession> cacheSession;
 
      // Open a stream based cache session.
    rv = cacheService->CreateSession("wyciwyg", nsICache::STORE_ANYWHERE, PR_TRUE, getter_AddRefs(cacheSession));
    if (!cacheSession) 
      return NS_ERROR_FAILURE;

    /* we'll try to synchronously open the cache entry... however, it may be
     * in use and not yet validated, in which case we'll try asynchronously
     * opening the cache entry.
     */
    if (aWriteAccess)
      rv = cacheSession->OpenCacheEntry(aCacheKey, nsICache::ACCESS_WRITE, PR_FALSE,
                                 getter_AddRefs(mCacheEntry));
    else
      rv = cacheSession->OpenCacheEntry(aCacheKey, nsICache::ACCESS_READ, PR_FALSE,
                                 getter_AddRefs(mCacheEntry));

    if (rv == NS_ERROR_CACHE_WAIT_FOR_VALIDATION) {
      // access to the cache entry has been denied. Let's try
      // opening it async. 
      rv = cacheSession->AsyncOpenCacheEntry(aCacheKey, nsICache::ACCESS_WRITE, this);
      if (NS_FAILED(rv)) 
        return rv;
      if (aDelayFlag)
        *aDelayFlag = PR_TRUE;
    }
    else if (rv == NS_OK) {
      LOG(("nsWyciwygChannel::OpenCacheEntry got cache entry \n"));
    }
  } 
  return rv;
}

nsresult
nsWyciwygChannel::Connect(PRBool firstTime)
{
  nsresult rv;

  LOG(("nsWyciwygChannel::Connect [this=%x]\n", this));

  // true when called from AsyncOpen
  if (firstTime) {
    PRBool delayed = PR_FALSE;

    nsXPIDLCString spec;
    mURI->GetSpec(getter_Copies(spec));
    // open a cache entry for this channel...
    rv = OpenCacheEntry(spec.get(), PR_FALSE, &delayed);        

    if (NS_FAILED(rv)) {
      LOG(("nsWyciwygChannel::Connect OpenCacheEntry failed [rv=%x]\n", rv));
      return rv;
    }
 
    if (NS_SUCCEEDED(rv) && delayed)
      return NS_OK;
  }

  // Read the script from cache. 
  if (mCacheEntry) 
    return ReadFromCache();
  return rv;
}

nsresult
nsWyciwygChannel::ReadFromCache()
{
    nsresult rv;

    NS_ENSURE_TRUE(mCacheEntry, NS_ERROR_FAILURE);

    LOG(("nsWyciwygChannel::ReadFromCache [this=%x] ", this));

    // Get a transport to the cached data...
    rv = mCacheEntry->GetTransport(getter_AddRefs(mCacheTransport));
    if (NS_FAILED(rv) || !mCacheTransport) 
      return rv;

    // Hookup the notification callbacks interface to the new transport...
    mCacheTransport->SetNotificationCallbacks(this, 
                                  ((mLoadFlags & nsIRequest::LOAD_BACKGROUND) 
                                    ? nsITransport::DONT_REPORT_PROGRESS 
                                    : 0));

    // Pump the cache data downstream
    return mCacheTransport->AsyncRead(this, nsnull,
                                      0, PRUint32(-1), 0,
                                      getter_AddRefs(mCacheReadRequest));
}




// called when Connect fails
nsresult
nsWyciwygChannel::AsyncAbort(nsresult aStatus)
{
    LOG(("nsWyciwygChannel::AsyncAbort [this=%x status=%x]\n", this, aStatus));

    mStatus = aStatus;
    mIsPending = PR_FALSE;

    // Remove ourselves from the load group.
    if (mLoadGroup)
        mLoadGroup->RemoveRequest((nsIRequest *)this, nsnull, aStatus);

    return NS_OK;
}

